#!/usr/bin/env node
import fs from 'fs';
import commander from 'commander';
import chalk from 'chalk';
import { __assign, __rest } from 'tslib';
import Ajv from 'ajv';
import 'ts-array-extensions';
import path from 'path';
import yaml from 'yaml';
import prettier from 'prettier';

/* eslint-disable no-console */
var o = process.stdout;
var e = process.stdout;
var progressLogger = function (start) {
    var last = 0;
    return function (m) {
        o.cursorTo(start);
        o.write(m);
        if (m.length < last) {
            o.write(' '.repeat(last - m.length));
        }
        last = m.length;
    };
};
var success = function () { return chalk.green(' -> SUCCESS'); };
var progress = function (message) {
    return chalk.blue(" -> " + message);
};
var processLogger = function () {
    var log = function (m) {
        o.write("\n" + m);
    };
    return {
        error: function (m) {
            e.write("\n" + chalk.red(m));
        },
        log: log,
        create: function (m) {
            log(m);
            var l = progressLogger(m.length);
            l(progress('...'));
            return l;
        }
    };
};

var initUpper = function (value) {
    return value.replace(/(?:^|\s)[a-z]/g, function (m) { return m.toUpperCase(); });
};

var dereference = function (references, $ref, prefix) {
    if (!$ref.startsWith(prefix)) {
        throw new Error("Invalid $ref " + $ref);
    }
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    var maybe = references === null || references === void 0 ? void 0 : references[name];
    if (!maybe)
        throw new Error("Failed to deference " + $ref);
    return maybe;
};

var dereferenceResponseObject = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference(doc.responses, $ref, '#/responses/');
};

var dereferenceParameterObject = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference(doc.parameters, $ref, '#/parameters/');
};

var mapProperties = function (properties) {
    var r = {};
    Object.keys(properties).forEach(function (key) {
        // eslint-disable-next-line @typescript-eslint/no-use-before-define
        r[key] = convertSchemaToJsonSchema(properties[key]);
    });
    return r;
};
var convertSchemaToJsonSchema = function (schema) {
    /**
     * This currently does nothing, but is a placeholder for removing
     * unsupported features from Open Api
     */
    var properties = schema.properties, rest = __rest(schema, ["properties"]);
    return __assign(__assign({}, rest), (properties ? { properties: mapProperties(properties) } : undefined));
};

function isReferenceObject(object) {
    if (typeof object !== 'object') {
        return false;
    }
    if (object === null) {
        return false;
    }
    return '$ref' in object;
}

var mapParameters = function (document) {
    return Object.keys(document.paths).flatMap(function (path) {
        if (typeof path !== 'string')
            return [];
        var pathObject = document.paths[path];
        return Object.keys(pathObject).compactMap(function (method) {
            var operationObject = pathObject[method];
            if (typeof operationObject !== 'object') {
                return undefined;
            }
            if (!('operationId' in operationObject) || !operationObject.operationId) {
                throw new Error("OperationId required for " + method + " " + path);
            }
            var operationId = operationObject.operationId;
            var parameters = operationObject.parameters;
            var inlined = parameters === null || parameters === void 0 ? void 0 : parameters.map(function (p) {
                return isReferenceObject(p) ? dereferenceParameterObject(p, document) : p;
            });
            return {
                path: path,
                method: method,
                operationId: operationId,
                operation: operationObject,
                parameters: inlined !== null && inlined !== void 0 ? inlined : []
            };
        });
    });
};

var fromV2$5 = function (document, log) {
    var inlineResponseBodyDefinitions = {};
    mapParameters(document).forEach(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.operation.responses;
        parameters === null || parameters === void 0 ? void 0 : parameters.forEach(function (parameterObject) {
            if (parameterObject.in !== 'body')
                return;
            var schema = parameterObject.schema;
            var key = initUpper(operationId) + "RequestBody";
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            inlineResponseBodyDefinitions[key] = convertSchemaToJsonSchema(schema);
        });
        Object.keys(responses).forEach(function (statusCode) {
            var response = responses[statusCode];
            var responseObject = isReferenceObject(response)
                ? dereferenceResponseObject(response, document)
                : response;
            if (!responseObject) {
                return;
            }
            var schema = responseObject.schema;
            if (!schema) {
                return;
            }
            var key = "" + initUpper(operationId) + statusCode + "ResponseBody";
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            inlineResponseBodyDefinitions[key] = convertSchemaToJsonSchema(schema);
        });
    });
    var documentDefinitions = {};
    Object.keys(document.definitions).forEach(function (key) {
        var _a;
        var value = (_a = document.definitions) === null || _a === void 0 ? void 0 : _a[key];
        if (value) {
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            documentDefinitions[key] = convertSchemaToJsonSchema(value);
        }
    });
    var out = {
        $schema: 'http://json-schema.org/draft-07/schema#',
        definitions: __assign(__assign({}, documentDefinitions), inlineResponseBodyDefinitions)
    };
    return out;
};
var fromV3$5 = function () {
    throw new Error('OpenApi 3 not supported');
};
var createJsonSchemaFromOpenApiDocument = function (_a) {
    var openApiDocument = _a.openApiDocument, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create("Generating JSON schema from Open API Document");
    var schema = 'swagger' in openApiDocument ? fromV2$5(openApiDocument, log) : fromV3$5();
    log === null || log === void 0 ? void 0 : log(progress("Validating JSON schema"));
    var ajv = new Ajv({ strict: false });
    ajv.addSchema(schema);
    Object.keys(schema.definitions).forEach(function (key) {
        ajv.getSchema("#/definitions/" + key);
    });
    log === null || log === void 0 ? void 0 : log(success());
    return schema;
};

var initLower = function (value) {
    return value.replace(/(?:^|\s)[A-Z]/g, function (m) { return m.toLowerCase(); });
};

var fromV2$4 = function (document, log) {
    return mapParameters(document).map(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.operation.responses;
        var paramsType = "Param." + initUpper(operationId) + "Params";
        var responseBodyType = Object.keys(responses)
            .compactMap(function (statusCode) {
            var response = responses[statusCode];
            var responseObject = isReferenceObject(response)
                ? dereferenceResponseObject(response, document)
                : response;
            if (!responseObject) {
                return undefined;
            }
            var schema = responseObject.schema;
            if (!schema) {
                return undefined;
            }
            return "Schema." + initUpper(operationId) + statusCode + "ResponseBody";
        })
            .join(' | ') || 'unknown';
        var statusCodes = Object.keys(responses)
            .map(function (k) { return (k === 'default' ? 'number' : k); })
            .join(' | ');
        var requestBodyType = parameters.some(function (p) { return p.in === 'body'; })
            ? "Schema." + initUpper(operationId) + "RequestBody"
            : 'unknown';
        var queryType = "Query." + initUpper(operationId) + "Query";
        var typeName = initUpper(operationId) + "RequestHandler";
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return {
            code: "export type " + typeName + " =\n          RequestHandler<\n            " + paramsType + ",\n            " + responseBodyType + ",\n            " + requestBodyType + ",\n            " + queryType + ",\n            Record<string, any>,\n            " + statusCodes + ">",
            operationId: operationId
        };
    });
};
var fromV3$4 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateRequestHandlersTypes = function (_a) {
    var jsonSchemaTypesModuleName = _a.jsonSchemaTypesModuleName, logger = _a.logger, openApiDocument = _a.openApiDocument, pathParameterTypesModuleName = _a.pathParameterTypesModuleName, queryTypesModuleName = _a.queryTypesModuleName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating typescript types for request handlers');
    var types = 'swagger' in openApiDocument ? fromV2$4(openApiDocument, log) : fromV3$4();
    var code = "\n  import { RequestHandler } from '@laurence79/express-async-request-handler';\n  import * as Schema from './" + jsonSchemaTypesModuleName + "';\n  import * as Query from './" + queryTypesModuleName + "';\n  import * as Param from './" + pathParameterTypesModuleName + "';\n  \n  " + types.map(function (t) { return t.code; }).join('\n\n') + "\n\n  export type RequestHandlers = {\n    " + types
        .map(function (t) {
        return initLower(t.operationId) + ": " + initUpper(t.operationId) + "RequestHandler";
    })
        .join(';\n') + "\n  };\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var validatorTemplate = function (operationIds, jsonSchemaFilename) { return "\nimport Ajv, { ValidateFunction } from 'ajv';\nimport { RequestHandler } from 'express';\nimport schema from './" + jsonSchemaFilename + "';\n\nconst ajv = new Ajv();\najv.addSchema(schema);\n\nconst bodyValidator = (validator: ValidateFunction): RequestHandler => {\n  return (req, res, next) => {\n    if (validator(req.body)) {\n      return next();\n    }\n  \n    res.status(400).send({\n      errors: validator.errors?.map(e => e.message ?? '') ?? [\n        'Unknown validation error'\n      ]\n    });\n  };\n}\n\n" + operationIds
    .map(function (id) {
    var typeName = initUpper(id);
    var validatorName = initLower(id);
    return "\nconst " + id + "Validator = ajv.getSchema('#/definitions/" + typeName + "RequestBody')!;\nexport const " + validatorName + " = bodyValidator(" + id + "Validator);\n\n";
})
    .join('') + "\n"; };

var pathParameterTypeName = function (operationId) {
    return initUpper(operationId) + "Params";
};

var fromV2$3 = function (document, log) {
    return mapParameters(document).map(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters;
        var typeName = pathParameterTypeName(operationId);
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        var fields = parameters
            .filter(function (p) { return p.in === 'path'; })
            .map(function (_a) {
            var name = _a.name;
            return name;
        });
        var definition = fields.none()
            ? 'ParamsDictionary'
            : "{ " + fields.map(function (field) { return field + ": string"; }).join('; ') + " }";
        return "export type " + typeName + " = " + definition;
    });
};
var fromV3$3 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generatePathParameterTypes = function (_a) {
    var openApiDocument = _a.openApiDocument, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating typescript types for path parameters');
    var types = 'swagger' in openApiDocument ? fromV2$3(openApiDocument, log) : fromV3$3();
    var code = "\n  import { ParamsDictionary } from 'express-serve-static-core';\n  \n  " + types.join('\n\n') + "\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var fromV2$2 = function (document, log) {
    return mapParameters(document).map(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters;
        var typeName = initUpper(operationId) + "Query";
        log === null || log === void 0 ? void 0 : log(progress("Adding " + typeName));
        var fields = parameters
            .filter(function (p) { return p.in === 'query'; })
            .map(function (_a) {
            var name = _a.name;
            return name;
        });
        var definition = fields.none()
            ? 'ParsedQs'
            : "{ " + fields.map(function (field) { return field + "?: string"; }).join('; ') + " }";
        return "export type " + typeName + " = " + definition;
    });
};
var fromV3$2 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateURLQueryTypes = function (_a) {
    var openApiDocument = _a.openApiDocument, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating typescript types for queries');
    var types = 'swagger' in openApiDocument ? fromV2$2(openApiDocument, log) : fromV3$2();
    var code = "\n  import { ParsedQs } from 'qs';\n  \n  " + types.join('\n\n') + "\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var fromV2$1 = function (document) {
    return mapParameters(document).compactMap(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters;
        if (parameters.some(function (p) { return p.in === 'body'; })) {
            return operationId;
        }
        return undefined;
    });
};
var fromV3$1 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateJSONSchemaValidators = function (_a) {
    var jsonSchemaFilename = _a.jsonSchemaFilename, openApiDocument = _a.openApiDocument, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating JSON schema validators');
    var operationIds = 'swagger' in openApiDocument ? fromV2$1(openApiDocument) : fromV3$1();
    var code = validatorTemplate(operationIds, jsonSchemaFilename);
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var fromV2 = function (document, log) {
    return mapParameters(document).map(function (_a) {
        var method = _a.method, parameters = _a.parameters, path = _a.path, operationId = _a.operationId;
        var hasBody = parameters.some(function (p) { return p.in === 'body'; });
        var bodyValidatorName = "Validators." + initLower(operationId);
        var handlerType = "RequestHandlers." + initUpper(operationId) + "RequestHandler";
        var handlerInstanceName = "" + initLower(operationId);
        var requestHandlersPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":" + x.substr(1, x.length - 2); });
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + requestHandlersPath));
        return {
            route: "." + method + "('" + requestHandlersPath + "', " + (hasBody ? bodyValidatorName + ", " : '') + " " + handlerInstanceName + ")",
            handlerInstanceName: handlerInstanceName,
            handlerType: handlerType
        };
    });
};
var fromV3 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateRouter = function (_a) {
    var jsonSchemaValidatorsModuleName = _a.jsonSchemaValidatorsModuleName, logger = _a.logger, openApiDocument = _a.openApiDocument, requestHandlersModuleName = _a.requestHandlersModuleName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating router');
    var types = 'swagger' in openApiDocument ? fromV2(openApiDocument, log) : fromV3();
    var argumentList = types
        .map(function (_a) {
        var handlerInstanceName = _a.handlerInstanceName;
        return handlerInstanceName;
    })
        .join(', ');
    var code = "\n  import { Router } from 'express';\n  import * as RequestHandlers from './" + requestHandlersModuleName + "';\n  import * as Validators from './" + jsonSchemaValidatorsModuleName + "';\n  \n  export const router = ({ " + argumentList + " }: RequestHandlers.RequestHandlers): Router => {\n    return Router()\n      " + types.map(function (t) { return t.route; }).join('\n') + "\n  }\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var typeDefForObject = function (objectSchema, recursiveLookup) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject);
        return "" + propertyName + (isRequired ? '' : '?') + ": " + propertyType;
    });
    return "{ " + propertyList.join('; ') + " }";
};

var typeDefForEnum = function (members) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"" + e + "\"";
        }
        if (e === null) {
            return 'null';
        }
        return e.toString();
    })
        .join(' | ');
};

var typeDefForReference = function ($ref) {
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    return name;
};

var typeDefForSchema = function (schema) {
    var _a;
    if (typeof schema === 'boolean') {
        throw new Error("Schema nodes must be an object.");
    }
    if ('$ref' in schema && schema.$ref) {
        return typeDefForReference(schema.$ref);
    }
    if (!('type' in schema)) {
        throw new Error("Schema nodes must have a type.");
    }
    var type = schema.type, enumProp = schema.enum;
    if (type === 'object') {
        return typeDefForObject(schema, typeDefForSchema);
    }
    if (enumProp) {
        return typeDefForEnum(enumProp);
    }
    if (type === 'boolean' ||
        type === 'null' ||
        type === 'number' ||
        type === 'string') {
        return type;
    }
    if (type === 'integer') {
        return 'number';
    }
    if (type === 'array') {
        var items = schema.items;
        if (!items) {
            throw new Error("Array nodes must have an items property.");
        }
        if (Array.isArray(items)) {
            throw new Error("Array items property must be singular.");
        }
        return "Array<" + typeDefForSchema(items) + ">";
    }
    throw new Error("Unrecognized type " + ((_a = type === null || type === void 0 ? void 0 : type.toString()) !== null && _a !== void 0 ? _a : ''));
};

var generateTypesFromJsonSchema = function (_a) {
    var jsonSchema = _a.jsonSchema, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating typescript types from JSON schema');
    var definitions = jsonSchema.definitions;
    if (!definitions)
        return '';
    var types = Object.keys(definitions).compactMap(function (typeName) {
        if (typeof typeName !== 'string')
            return undefined;
        var definition = definitions[typeName];
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return "export type " + typeName + " = " + typeDefForSchema(definition) + ";";
    });
    var code = types.join('\n\n');
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var loadOpenApiDocument = function (_a) {
    var openApiDocumentFilename = _a.openApiDocumentFilename, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create("Loading swagger file " + openApiDocumentFilename);
    if (!fs.existsSync(openApiDocumentFilename)) {
        throw new Error("File " + openApiDocumentFilename + " does not exist");
    }
    var extension = path.extname(openApiDocumentFilename);
    var data = fs.readFileSync(openApiDocumentFilename, 'UTF-8');
    var doc = (function () {
        if (['.yaml', '.yml'].includes(extension)) {
            return yaml.parse(data);
        }
        if (extension === '.json') {
            return JSON.parse(data);
        }
        throw new Error("Unsupported extension " + extension);
    })();
    log === null || log === void 0 ? void 0 : log(success());
    return doc;
};

var prettierStyle = {
    arrowParens: 'avoid',
    semi: true,
    trailingComma: 'none',
    singleQuote: true,
    printWidth: 80
};
var writeFiles = function (_a) {
    var files = _a.files, outputDirectory = _a.outputDirectory, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Writing files');
    log === null || log === void 0 ? void 0 : log(progress('Creating output directory'));
    fs.mkdirSync(outputDirectory, { recursive: true });
    files.forEach(function (_a) {
        var filename = _a.filename, content = _a.content;
        log === null || log === void 0 ? void 0 : log(progress(filename));
        var extension = path.extname(filename);
        var parser = (function () {
            switch (extension) {
                case '.ts':
                    return 'typescript';
                case '.json':
                    return 'json';
                default:
                    return undefined;
            }
        })();
        var fullFileName = path.join(outputDirectory, filename);
        var formattedContent = prettier.format(content, __assign(__assign({}, prettierStyle), { parser: parser }));
        fs.writeFileSync(fullFileName, formattedContent);
    });
    log === null || log === void 0 ? void 0 : log(success());
};

var defaultOptions = {
    jsonSchemaFilename: 'schema.json',
    jsonSchemaTypesModuleName: 'types',
    jsonSchemaValidatorsModuleName: 'validators',
    outputDirectory: 'generated',
    pathParameterTypesModuleName: 'parameterTypes',
    requestHandlersModuleName: 'handlers',
    queryTypesModuleName: 'queryTypes',
    routerModuleName: 'router'
};
var generateCode = function (inputOptions) {
    var options = __assign(__assign({}, defaultOptions), inputOptions);
    var openApiDocument = loadOpenApiDocument(options);
    var jsonSchema = createJsonSchemaFromOpenApiDocument(__assign(__assign({}, options), { openApiDocument: openApiDocument }));
    var renderDeps = __assign(__assign({}, options), { openApiDocument: openApiDocument,
        jsonSchema: jsonSchema });
    var files = [
        {
            filename: options.jsonSchemaFilename,
            content: JSON.stringify(jsonSchema)
        },
        {
            filename: options.jsonSchemaTypesModuleName + ".ts",
            content: generateTypesFromJsonSchema(renderDeps)
        },
        {
            filename: options.jsonSchemaValidatorsModuleName + ".ts",
            content: generateJSONSchemaValidators(renderDeps)
        },
        {
            filename: options.queryTypesModuleName + ".ts",
            content: generateURLQueryTypes(renderDeps)
        },
        {
            filename: options.pathParameterTypesModuleName + ".ts",
            content: generatePathParameterTypes(renderDeps)
        },
        {
            filename: options.requestHandlersModuleName + ".ts",
            content: generateRequestHandlersTypes(renderDeps)
        },
        {
            filename: options.routerModuleName + ".ts",
            content: generateRouter(renderDeps)
        }
    ];
    writeFiles(__assign(__assign({}, options), { files: files }));
};

var packageJsonContent = fs.readFileSync(__dirname + "/../package.json", 'utf-8');
var packageJson = JSON.parse(packageJsonContent);
var version = packageJson.version;
var cmd = new commander.Command('generate')
    .description('Generates Express types and functions from the spec.')
    .arguments('<spec-filename> <output-directory>')
    .action(function (specFilename, outputDirectory) {
    var logger = processLogger();
    try {
        generateCode({
            openApiDocumentFilename: specFilename,
            outputDirectory: outputDirectory,
            logger: logger
        });
        logger.log(chalk.green('\nCompleted successfully!\n\n'));
    }
    catch (e) {
        logger.error("\nFAILED. " + (e instanceof Error ? e.message : 'Unknown error') + "\n\n");
    }
});
commander.program.version(version).addCommand(cmd).parse();

export { generateCode };
